REM
REM  Emulation core of the Zilog 80 CPU clone used in GameBoy consoles
REM  Part of the ZilogBoy project
REM  Copyright (C) Mateusz Viste 2010
REM

TYPE CpuRegisters ' Note: IX and IY have been removed from the GB Z80 clone
  REM 8 bit registers
  A AS UBYTE   ' Accumulator
  B AS UBYTE   ' Commonly used as a 8bit counter
  C AS UBYTE
  D AS UBYTE
  E AS UBYTE
  F AS UBYTE   ' Flags
  H AS UBYTE
  L AS UBYTE
  I AS UBYTE   ' Interrupt vector register
  R AS UBYTE   ' Refresh register / might be used to generate random numbers
  REM 16 bit register - Note: to write/read composite register, use dedicated functions! (eg. ReadRegAF)
  PC AS INTEGER ' Program counter (holds memory adress currently executed code). No function can change PC
  SP AS INTEGER ' The stack pointer. Holds current address of the top of the stack
END TYPE

DIM SHARED Register AS CpuRegisters
DIM SHARED InterruptsState AS BYTE = 1  '  1 = ON  0 = OFF   (this is the IME register)
DIM SHARED HaltState AS BYTE = 0      ' 0 = CPU is running   1 = CPU is HALTed (wait until next interrupt)
DIM SHARED CpuInstruction AS INTEGER ' holds current CPU instruction
DIM SHARED cpc(0 TO 255) AS UBYTE    ' number of cycles to waste for every CPU instruction


#INCLUDE ONCE "cpu-timer.bi"      ' Include timer emulation
#INCLUDE ONCE "cpu-microcode.bi"  ' add microinstruction set


SUB CheckInterrupts()
  ' IME: Interrupt Master Enable (InterruptsState)
  ' IE: Interrupt Enable (&hFFFF)
  ' IF: Interrupt flag   (&hFF0F)
  DIM TempState AS UINTEGER
  TempState = MemoryRead(&hFFFF) AND MemoryRead(&hFF0F)    ' IE AND IF
  IF TempState > 0 AND InterruptsState = 1 THEN   ' Check also InterruptsState (IME)
    IF HaltState = 1 THEN HaltState = 0 : Register.PC += 1    ' If the CPU was HALTed, resume it
    PushToStack(Register.PC)
    InterruptsState = 0
    IF TempState AND &b00000001 = &b00000001 THEN  ' V-Blank
      Register.PC = 64
      IoRegisters(&hFF0F) = IoRegisters(&hFF0F) AND &b11111110  ' Reset the trigerred IF flag
    ELSEIF TempState AND &b00000010 = &b00000010 THEN  ' LCDC
      Register.PC = 72
      IoRegisters(&hFF0F) = IoRegisters(&hFF0F) AND &b11111101  ' Reset the trigerred IF flag
    ELSEIF TempState AND &b00000100 = &b00000100 THEN  ' TIMER
      Register.PC = TIMER
      IoRegisters(&hFF0F) = IoRegisters(&hFF0F) AND &b11111011  ' Reset the trigerred IF flag
    ELSEIF TempState AND &b00001000 = &b00001000 THEN  ' SERIAL
      Register.PC = 88
      IoRegisters(&hFF0F) = IoRegisters(&hFF0F) AND &b11110111  ' Reset the trigerred IF flag
    ELSEIF TempState AND &b00010000 = &b00010000 THEN ' Joypad
      Register.PC = 96
      IoRegisters(&hFF0F) = IoRegisters(&hFF0F) AND &b11101111  ' Reset the trigerred IF flag
    END IF
  END IF
END SUB


SUB CpuExec()
  DIM AS STRING LastKey
  DIM AS INTEGER OldLine
  DIM AS UBYTE UbyteBuff   ' Temporary buffer for various UBYTE values
  #IFDEF DEBUGMODE
    IF CSRLIN() >= 111 THEN LOCATE 1,50 ELSE LOCATE ,50  ' DEBUG
    PRINT "                                    ";        ' DEBUG
    LOCATE ,50                                           ' DEBUG
    PRINT HEX(Register.PC, 4); ": ";                     ' DEBUG
  #ENDIF
  SELECT CASE MemoryRead(Register.PC)  ' Check if the instruction is 16 bits
    CASE &hCB, &h10
      CpuInstruction = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC))
    CASE ELSE  ' Instruction is 8 bit
      CpuInstruction = MemoryRead(Register.PC)
  END SELECT

  SELECT CASE CpuInstruction
    CASE &h00   ' NOP (No operation)
      PrintDebug("NOP")  ' DEBUG
      Register.PC += 1
    CASE &h01   ' LD BC,nn  (load nn into BC)
      PrintDebug("LD BC," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))))  ' DEBUG
      WriteRegBC(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))
      Register.PC += 3
    CASE &h02   ' LD (BC),A   ' Load value of A into address at BC
      PrintDebug("LD (BC),A")  ' DEBUG
      MemoryWrite(ReadRegBC(), Register.A)
      Register.PC += 1
    CASE &h03   ' INC BC (register BC+=1)
      PrintDebug("INC BC")  ' DEBUG
      IF ReadRegBC() = 65535 THEN WriteRegBC(0) ELSE WriteRegBC(ReadRegBC() + 1)
      Register.PC += 1
    CASE &h04   ' INC B (register B+=1)
      PrintDebug("INC B")  ' DEBUG
      IncReg8(Register.B)
      Register.PC += 1
    CASE &h05   ' DEC B (register B-=1)
      PrintDebug("DEC B")  ' DEBUG
      DecReg8(Register.B)
      Register.PC += 1
    CASE &h06   ' LD B,n (load n into B)
      PrintDebug("LD B," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.B = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h07   ' RLCA (Rotate reg A left, old bit 7 written to Carry flag)
      PrintDebug("RLCA")  '  DEBUG
      RLCA()
      Register.PC += 1
    CASE &h08   ' LD (nn),SP  (puts SP value at nn)
      PrintDebug("LD " & HEX(DwordVal(MemoryRead(Register.PC + 1),MemoryRead(Register.PC + 2))) & ",SP")
      MemoryWrite16(DwordVal(MemoryRead(Register.PC + 1),MemoryRead(Register.PC + 2)), Register.SP)
      Register.PC += 3
    CASE &h09   ' ADD HL,BC
      PrintDebug("ADD HL,BC")
      AddToHL(ReadRegBC())
      Register.PC += 1
    CASE &h0A   ' LD A,(BC)  (put value from address BC to A)
      PrintDebug("LD A,(BC)")  ' DEBUG
      Register.A = MemoryRead(ReadRegBC())
      Register.PC += 1
    CASE &h0B   ' DEC BC (register BC-=1)
      PrintDebug("DEC BC")  ' DEBUG
      IF ReadRegBC() = 0 THEN WriteRegBC(65535) ELSE WriteRegBC(ReadRegBC() - 1)
      Register.PC += 1
    CASE &h0C   ' INC C (register C+=1)
      PrintDebug("INC C")  ' DEBUG
      IncReg8(Register.C)
      Register.PC += 1
    CASE &h0D   ' DEC C (register C-=1)
      PrintDebug("DEC C")  ' DEBUG
      DecReg8(Register.C)
      Register.PC += 1
    CASE &h0E   ' LD C,n (load n into C)
      PrintDebug("LD C," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.C = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h0F   ' RRCA (Rotate reg A right, old bit 0 written to Carry flag)
      PrintDebug("RRCA")  '  DEBUG
      RRCA()
      Register.PC += 1
    CASE &h11   ' LD DE,nn  (load nn into DE)
      PrintDebug("LD DE," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))))  ' DEBUG
      WriteRegDE(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))
      Register.PC += 3
    CASE &h12   ' LD (DE),A   ' Load value of A into address at DE
      PrintDebug("LD (DE),A")  ' DEBUG
      MemoryWrite(ReadRegDE(), Register.A)
      Register.PC += 1
    CASE &h13   ' INC DE (register DE+=1)
      PrintDebug("INC DE")  ' DEBUG
      IF ReadRegDE() = 65535 THEN WriteRegDE(0) ELSE WriteRegDE(ReadRegDE() + 1)
      Register.PC += 1
    CASE &h14   ' INC D (register D+=1)
      PrintDebug("INC D")  ' DEBUG
      IncReg8(Register.D)
      Register.PC += 1
    CASE &h15   ' DEC D (register D-=1)
      PrintDebug("DEC D")  ' DEBUG
      DecReg8(Register.D)
      Register.PC += 1
    CASE &h16   ' LD D,n (load n into D)
      PrintDebug("LD D," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.D = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h17  ' RLA
      PrintDebug("RLA")  ' DEBUG
      RotateLeftCarry(Register.A)
      Register.PC += 1
    CASE &h18   ' JR n (jump to PC+n)
      PrintDebug("JR " & UbyteToByte(MemoryRead(Register.PC + 1)))  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then jump
      Register.PC += UbyteToByte(MemoryRead(Register.PC - 1))
    CASE &h19   ' ADD HL,DE
      PrintDebug("ADD HL,DE")
      AddToHL(ReadRegDE())
      Register.PC += 1
    CASE &h1A   ' LD A,(DE)  (put value from address DE to A)
      PrintDebug("LD A,(DE)")
      Register.A = MemoryRead(ReadRegDE())
      Register.PC += 1
    CASE &h1B   ' DEC DE (register DE-=1)
      PrintDebug("DEC DE")  ' DEBUG
      IF ReadRegDE() = 0 THEN WriteRegDE(65535) ELSE WriteRegDE(ReadRegDE() - 1)
      Register.PC += 1
    CASE &h1C   ' INC E (register E+=1)
      PrintDebug("INC E")  ' DEBUG
      IncReg8(Register.E)
      Register.PC += 1
    CASE &h1D   ' DEC E (register E-=1)
      PrintDebug("DEC E")  ' DEBUG
      DecReg8(Register.E)
      Register.PC += 1
    CASE &h1E   ' LD E,n (load n into E)
      PrintDebug("LD E," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.E = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h1F  ' RRA
      PrintDebug("RRA")  ' DEBUG
      RotateRightCarry(Register.A)
      Register.PC += 1
    CASE &h20   ' JR NZ,n (jump to PC+n if Z flag is false)
      PrintDebug("JR NZ," & UbyteToByte(MemoryRead(Register.PC + 1)))  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagZ() = 0 THEN Register.PC += UbyteToByte(MemoryRead(Register.PC - 1))
    CASE &h21   ' LD HL,nn  (load nn into HL)
      PrintDebug("LD HL," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))))  ' DEBUG
      WriteRegHL(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))
      Register.PC += 3
    CASE &h22   ' LD (HL+),A (put A into memory addr HL, and increments HL)
      PrintDebug("LD (HL+),A")   ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      IF ReadRegHL() = 65535 THEN WriteRegHL(0) ELSE WriteRegHL(ReadRegHL() + 1)
      Register.PC += 1
    CASE &h23   ' INC HL (register HL+=1)
      PrintDebug("INC HL")  ' DEBUG
      IF ReadRegHL() = 65535 THEN WriteRegHL(0) ELSE WriteRegHL(ReadRegHL() + 1)
      Register.PC += 1
    CASE &h24   ' INC H (register H+=1)
      PrintDebug("INC H")  ' DEBUG
      IncReg8(Register.H)
      Register.PC += 1
    CASE &h25   ' DEC H (register H-=1)
      PrintDebug("DEC H")  ' DEBUG
      DecReg8(Register.H)
      Register.PC += 1
    CASE &h26   ' LD H,n (load n into H)
      PrintDebug("LD H," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.H = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h27   ' DAA
      PrintDebug("DAA")  ' DEBUG
      AdjustDAA()
      Register.PC += 1
    CASE &h29   ' ADD HL,HL
      PrintDebug("ADD HL,HL")
      AddToHL(ReadRegHL())
      Register.PC += 1
    CASE &h2A   ' LD A,(HL+) (Put value @HL into A, and increment HL by 1)
      PrintDebug("LD A,(HL+)")    ' DEBUG
      Register.A = MemoryRead(ReadRegHL())
      IF ReadRegHL() = 65535 THEN WriteRegHL(0) ELSE WriteRegHL(ReadRegHL() + 1)
      Register.PC += 1
    CASE &h2B   ' DEC HL (register HL-=1)
      PrintDebug("DEC HL")  ' DEBUG
      IF ReadRegHL() = 0 THEN WriteRegHL(65535) ELSE WriteRegHL(ReadRegHL() - 1)
      Register.PC += 1
    CASE &h2C   ' INC L (register L+=1)
      PrintDebug("INC L")  ' DEBUG
      IncReg8(Register.L)
      Register.PC += 1
    CASE &h2D   ' DEC L (register L-=1)
      PrintDebug("DEC L")  ' DEBUG
      DecReg8(Register.L)
      Register.PC += 1
    CASE &h2E   ' LD L,n (load n into L)
      PrintDebug("LD L," & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      Register.L = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h28   ' JR Z,n (jump to PC+n if Z flag is true)
      PrintDebug("JR Z," & UbyteToByte(MemoryRead(Register.PC + 1)))  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagZ() = 1 THEN Register.PC += UbyteToByte(MemoryRead(Register.PC - 1))
    CASE &h2F   ' CPL (complements the A register, that is flips all bits)
      PrintDebug("CPL")
      SetFlagN() : SetFlagH()
      Register.A = Register.A XOR &b11111111  ' Flip all bits
      Register.PC += 1
    CASE &h30   ' JR NC,n (jump to PC+n if C flag is false)
      PrintDebug("JR NC," & UbyteToByte(MemoryRead(Register.PC + 1)))  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagC() = 0 THEN Register.PC += UbyteToByte(MemoryRead(Register.PC - 1))
    CASE &h31   ' LD SP,nn
      PrintDebug("LD SP," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))))  ' DEBUG
      Register.SP = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
      Register.PC += 3
    CASE &h32   ' LD (HL-),A   (put A into address at HL, then decrement HL)
      PrintDebug("LD (HL-),A")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      IF ReadRegHL() = 0 THEN WriteRegHL(65535) ELSE WriteRegHL(ReadRegHL() - 1)
      Register.PC += 1
    CASE &h33   ' INC SP (register SP+=1)
      PrintDebug("INC SP")  ' DEBUG
      IF Register.SP = 65535 THEN Register.SP = 0 ELSE Register.SP += 1
      Register.PC += 1
    CASE &h34   ' INC HL (value at (HL) is incremented)
      PrintDebug("INC (HL)")   ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      IncReg8(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 1
    CASE &h35   ' DEC HL (HL-=1)
      PrintDebug("DEC (HL)")   ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())  ' Save old value at (HL)
      DecReg8(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 1
    CASE &h36   ' LD (HL),n   ' Load value of n into address at HL
      PrintDebug("LD (HL)," & HEX(MemoryRead(Register.PC + 1),2))  ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &h37   ' SCF (Set Carry Flag)
      PrintDebug("SCF")  ' DEBUG
      ResetFlagN() : ResetFlagH() : SetFlagC()
      Register.PC += 1
    CASE &h38   ' JR C,n (jump to PC+n if C flag is true)
      PrintDebug("JR C," & UbyteToByte(MemoryRead(Register.PC + 1)))  ' DEBUG
      Register.PC += 2  ' First increment the PC, and then we'll see
      IF GetFlagC() = 1 THEN Register.PC += UbyteToByte(MemoryRead(Register.PC - 1))
    CASE &h39   ' ADD HL,SP
      PrintDebug("ADD HL,SP")
      AddToHL(Register.SP)
      Register.PC += 1
    CASE &h3A   ' LD A,(HL-)   (put value at address HL into A, and decrement HL)
      PrintDebug("LD A,(HL-)")
      Register.A = MemoryRead(ReadRegHL())
      IF ReadRegHL() = 0 THEN WriteRegHL(65535) ELSE WriteRegHL(ReadRegHL() - 1)
      Register.PC += 1
    CASE &h3B   ' DEC SP (register SP-=1)
      PrintDebug("DEC SP")  ' DEBUG
      IF Register.SP = 0 THEN Register.SP = 65535 ELSE Register.SP -= 1
      Register.PC += 1
    CASE &h3C   ' INC A (register A+=1)
      PrintDebug("INC A")  ' DEBUG
      IncReg8(Register.A)
      Register.PC += 1
    CASE &h3D   ' DEC A (register A-=1)
      PrintDebug("DEC A")  ' DEBUG
      DecReg8(Register.A)
      Register.PC += 1
    CASE &h3E   ' LD A,n  (put n into A)
      PrintDebug("LD A," & HEX(MemoryRead(Register.PC + 1),2))
      Register.A = MemoryRead(Register.PC + 1)
      Register.PC += 2
    CASE &h3F   ' CCF  (Complement Carry Flag)
      PrintDebug("CCF")
      ResetFlagN() : ResetFlagH()
      IF GetFlagC() = 0 THEN SetFlagC() ELSE ResetFlagC()
      Register.PC += 1
    CASE &h40   ' LD B,B     ' Put value B into B
      PrintDebug("LD B,B")  ' DEBUG
      Register.B = Register.B
      Register.PC += 1
    CASE &h41   ' LD B,C     ' Put value C into B
      PrintDebug("LD B,C")  ' DEBUG
      Register.B = Register.C
      Register.PC += 1
    CASE &h42   ' LD B,D     ' Put value D into B
      PrintDebug("LD B,D")  ' DEBUG
      Register.B = Register.D
      Register.PC += 1
    CASE &h43   ' LD B,E     ' Put value E into B
      PrintDebug("LD B,E")  ' DEBUG
      Register.B = Register.E
      Register.PC += 1
    CASE &h44   ' LD B,H     ' Put value H into B
      PrintDebug("LD B,H")  ' DEBUG
      Register.B = Register.H
      Register.PC += 1
    CASE &h45   ' LD B,L     ' Put value L into B
      PrintDebug("LD B,L")  ' DEBUG
      Register.B = Register.L
      Register.PC += 1
    CASE &h46   ' LD B,(HL)  (put value from address HL to B)
      PrintDebug("LD B,(HL)")
      Register.B = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h47   ' LD B,A     ' Put value A into B
      PrintDebug("LD B,A")  ' DEBUG
      Register.B = Register.A
      Register.PC += 1
    CASE &h48   ' LD C,B     ' Put value B into C
      PrintDebug("LD C,B")  ' DEBUG
      Register.C = Register.B
      Register.PC += 1
    CASE &h49   ' LD C,C     ' Put value C into C
      PrintDebug("LD C,C")  ' DEBUG
      Register.C = Register.C
      Register.PC += 1
    CASE &h4A   ' LD C,D     ' Put value D into C
      PrintDebug("LD C,D")  ' DEBUG
      Register.C = Register.D
      Register.PC += 1
    CASE &h4B   ' LD C,E     ' Put value E into C
      PrintDebug("LD C,E")  ' DEBUG
      Register.C = Register.E
      Register.PC += 1
    CASE &h4C   ' LD C,H     ' Put value H into C
      PrintDebug("LD C,H")  ' DEBUG
      Register.C = Register.H
      Register.PC += 1
    CASE &h4D   ' LD C,L     ' Put value L into C
      PrintDebug("LD C,L")  ' DEBUG
      Register.C = Register.L
      Register.PC += 1
    CASE &h4E   ' LD C,(HL)  (put value from address HL to C)
      PrintDebug("LD C,(HL)")
      Register.C = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h4F   ' LD C,A     ' Put value A into C
      PrintDebug("LD C,A")  ' DEBUG
      Register.C = Register.A
      Register.PC += 1
    CASE &h50   ' LD D,B     ' Put value B into D
      PrintDebug("LD D,B")  ' DEBUG
      Register.D = Register.B
      Register.PC += 1
    CASE &h51   ' LD D,C     ' Put value C into D
      PrintDebug("LD D,C")  ' DEBUG
      Register.D = Register.C
      Register.PC += 1
    CASE &h52   ' LD D,D     ' Put value D into D
      PrintDebug("LD D,D")  ' DEBUG
      Register.D = Register.D
      Register.PC += 1
    CASE &h53   ' LD D,E     ' Put value E into D
      PrintDebug("LD D,E")  ' DEBUG
      Register.D = Register.E
      Register.PC += 1
    CASE &h54   ' LD D,H     ' Put value H into D
      PrintDebug("LD D,H")  ' DEBUG
      Register.D = Register.H
      Register.PC += 1
    CASE &h55   ' LD D,L     ' Put value L into D
      PrintDebug("LD D,L")  ' DEBUG
      Register.D = Register.L
      Register.PC += 1
    CASE &h56   ' LD D,(HL)  (put value from address HL to D)
      PrintDebug("LD D,(HL)")
      Register.D = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h57   ' LD D,A     ' Put value A into D
      PrintDebug("LD D,A")  ' DEBUG
      Register.D = Register.A
      Register.PC += 1
    CASE &h58   ' LD E,B     ' Put value B into E
      PrintDebug("LD E,B")  ' DEBUG
      Register.E = Register.B
      Register.PC += 1
    CASE &h59   ' LD E,C     ' Put value C into E
      PrintDebug("LD E,C")  ' DEBUG
      Register.E = Register.C
      Register.PC += 1
    CASE &h5A   ' LD E,D     ' Put value D into E
      PrintDebug("LD E,D")  ' DEBUG
      Register.E = Register.D
      Register.PC += 1
    CASE &h5B   ' LD E,E     ' Put value E into E
      PrintDebug("LD E,E")  ' DEBUG
      Register.E = Register.E
      Register.PC += 1
    CASE &h5C   ' LD E,H     ' Put value H into E
      PrintDebug("LD E,H")  ' DEBUG
      Register.E = Register.H
      Register.PC += 1
    CASE &h5D   ' LD E,L     ' Put value L into E
      PrintDebug("LD E,L")  ' DEBUG
      Register.E = Register.L
      Register.PC += 1
    CASE &h5E   ' LD E,(HL)  (put value from address HL to E)
      PrintDebug("LD E,(HL)")
      Register.E = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h5F   ' LD E,A     ' Put value A into E
      PrintDebug("LD E,A")  ' DEBUG
      Register.E = Register.A
      Register.PC += 1
    CASE &h60   ' LD H,B     ' Put value B into H
      PrintDebug("LD H,B")  ' DEBUG
      Register.H = Register.B
      Register.PC += 1
    CASE &h61   ' LD H,C     ' Put value C into H
      PrintDebug("LD H,C")  ' DEBUG
      Register.H = Register.C
      Register.PC += 1
    CASE &h62   ' LD H,D     ' Put value D into H
      PrintDebug("LD H,D")  ' DEBUG
      Register.H = Register.D
      Register.PC += 1
    CASE &h63   ' LD H,E     ' Put value E into H
      PrintDebug("LD H,E")  ' DEBUG
      Register.H = Register.E
      Register.PC += 1
    CASE &h64   ' LD H,H     ' Put value H into H
      PrintDebug("LD H,H")  ' DEBUG
      'Register.H = Register.H   ' Commented out, because there is no point in doing that
      Register.PC += 1
    CASE &h65   ' LD H,L     ' Put value L into H
      PrintDebug("LD H,L")  ' DEBUG
      Register.H = Register.L
      Register.PC += 1
    CASE &h66   ' LD H,(HL)  (put value from address HL to H)
      PrintDebug("LD H,(HL)")
      Register.H = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h67   ' LD H,A     ' Put value A into H
      PrintDebug("LD H,A")  ' DEBUG
      Register.H = Register.A
      Register.PC += 1
    CASE &h68   ' LD L,B     ' Put value B into L
      PrintDebug("LD L,B")  ' DEBUG
      Register.L = Register.B
      Register.PC += 1
    CASE &h69   ' LD L,C     ' Put value C into L
      PrintDebug("LD L,C")  ' DEBUG
      Register.L = Register.C
      Register.PC += 1
    CASE &h6A   ' LD L,D     ' Put value D into L
      PrintDebug("LD L,D")  ' DEBUG
      Register.L = Register.D
      Register.PC += 1
    CASE &h6B   ' LD L,E     ' Put value E into L
      PrintDebug("LD L,E")  ' DEBUG
      Register.L = Register.E
      Register.PC += 1
    CASE &h6C   ' LD L,H     ' Put value H into L
      PrintDebug("LD L,H")  ' DEBUG
      Register.L = Register.H
      Register.PC += 1
    CASE &h6D   ' LD L,L     ' Put value L into L
      PrintDebug("LD L,L")  ' DEBUG
      'Register.L = Register.L   ' Commented out, because there is no point in doing such operation
      Register.PC += 1
    CASE &h6E   ' LD L,(HL)  (put value from address HL to L)
      PrintDebug("LD L,(HL)")
      Register.L = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h6F   ' LD L,A     ' Put value A into L
      PrintDebug("LD L,A")  ' DEBUG
      Register.L = Register.A
      Register.PC += 1
    CASE &h70   ' LD (HL),B   ' Load value of B into address at HL
      PrintDebug("LD (HL),B")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.B)
      Register.PC += 1
    CASE &h71   ' LD (HL),C   ' Load value of C into address at HL
      PrintDebug("LD (HL),C")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.C)
      Register.PC += 1
    CASE &h72   ' LD (HL),D   ' Load value of D into address at HL
      PrintDebug("LD (HL),D")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.D)
      Register.PC += 1
    CASE &h73   ' LD (HL),E   ' Load value of E into address at HL
      PrintDebug("LD (HL),E")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.E)
      Register.PC += 1
    CASE &h74   ' LD (HL),H   ' Load value of H into address at HL
      PrintDebug("LD (HL),H")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.H)
      Register.PC += 1
    CASE &h75   ' LD (HL),L   ' Load value of L into address at HL
      PrintDebug("LD (HL),L")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.L)
      Register.PC += 1
    CASE &h76   ' HALT   ' HALTs the CPU (only if interrupts are permitted) until next interrupt
      REM  Note: I don't increment PC here, to make the game loop on the HALT op.
      REM        PC will be incremented if an interrupt occurs.
      PrintDebug("Halt")
      IF InterruptsState = 1 THEN HaltState = 1
    CASE &h77   ' LD (HL),A   ' Load value of A into address at HL
      PrintDebug("LD (HL),A")  ' DEBUG
      MemoryWrite(ReadRegHL(), Register.A)
      Register.PC += 1
    CASE &h78   ' LD A,B  (put B into A)
      PrintDebug("LD A,B")   ' DEBUG
      Register.A = Register.B
      Register.PC += 1
    CASE &h79   ' LD A,C  (put C into A)
      PrintDebug("LD A,C")   ' DEBUG
      Register.A = Register.C
      Register.PC += 1
    CASE &h7A   ' LD A,D  (put D into A)
      PrintDebug("LD A,D")   ' DEBUG
      Register.A = Register.D
      Register.PC += 1
    CASE &h7B   ' LD A,E  (put E into A)
      PrintDebug("LD A,E")   ' DEBUG
      Register.A = Register.E
      Register.PC += 1
    CASE &h7C   ' LD A,H  (put H into A)
      PrintDebug("LD A,H")   ' DEBUG
      Register.A = Register.H
      Register.PC += 1
    CASE &h7D   ' LD A,L  (put L into A)
      PrintDebug("LD A,L")   ' DEBUG
      Register.A = Register.L
      Register.PC += 1
    CASE &h7E   ' LD A,(HL)  (put value from address HL to A)
      PrintDebug("LD A,(HL)")
      Register.A = MemoryRead(ReadRegHL())
      Register.PC += 1
    CASE &h7F   ' LD A,A  (put A into A)
      PrintDebug("LD A,A")   ' DEBUG
      'Register.A = Register.A   ' Commented out, because there is no point in doing such operation
      Register.PC += 1
    CASE &h80  ' ADD A,B  (A+=B)
      PrintDebug("ADD A,B")
      AddToA(Register.B)
      Register.PC += 1
    CASE &h81  ' ADD A,C  (A+=C)
      PrintDebug("ADD A,C")
      AddToA(Register.C)
      Register.PC += 1
    CASE &h82  ' ADD A,D  (A+=D)
      PrintDebug("ADD A,D")
      AddToA(Register.D)
      Register.PC += 1
    CASE &h83  ' ADD A,E  (A+=E)
      PrintDebug("ADD A,E")
      AddToA(Register.E)
      Register.PC += 1
    CASE &h84  ' ADD A,H  (A+=H)
      PrintDebug("ADD A,H")
      AddToA(Register.H)
      Register.PC += 1
    CASE &h85  ' ADD A,L  (A+=L)
      PrintDebug("ADD A,L")
      AddToA(Register.L)
      Register.PC += 1
    CASE &h86  ' ADD A,(HL)  (A+=[HL])
      PrintDebug("ADD A,(HL)")
      AddToA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &h87  ' ADD A,A  (A+=A)
      PrintDebug("ADD A,A")
      AddToA(Register.A)
      Register.PC += 1
    CASE &h88   ' ADC A,B
      PrintDebug("ADC A,B")   ' DEBUG
      AdcA(Register.B)
      Register.PC += 1
    CASE &h89   ' ADC A,C
      PrintDebug("ADC A,C")   ' DEBUG
      AdcA(Register.C)
      Register.PC += 1
    CASE &h8A   ' ADC A,D
      PrintDebug("ADC A,D")   ' DEBUG
      AdcA(Register.D)
      Register.PC += 1
    CASE &h8B   ' ADC A,E
      PrintDebug("ADC A,E")   ' DEBUG
      AdcA(Register.E)
      Register.PC += 1
    CASE &h8C   ' ADC A,H
      PrintDebug("ADC A,H")   ' DEBUG
      AdcA(Register.H)
      Register.PC += 1
    CASE &h8D   ' ADC A,L
      PrintDebug("ADC A,L")   ' DEBUG
      AdcA(Register.L)
      Register.PC += 1
    CASE &h8E   ' ADC A,(HL)
      PrintDebug("ADC A,(HL)")   ' DEBUG
      AdcA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &h8F   ' ADC A,A
      PrintDebug("ADC A,A")   ' DEBUG
      AdcA(Register.A)
      Register.PC += 1
    CASE &h90  ' SUB B
      PrintDebug("SUB B")  ' DEBUG
      SubValFromReg8(Register.A, Register.B)
      Register.PC += 1
    CASE &h91  ' SUB C
      PrintDebug("SUB C")  ' DEBUG
      SubValFromReg8(Register.A, Register.C)
      Register.PC += 1
    CASE &h92  ' SUB D
      PrintDebug("SUB D")  ' DEBUG
      SubValFromReg8(Register.A, Register.D)
      Register.PC += 1
    CASE &h93  ' SUB E
      PrintDebug("SUB E")  ' DEBUG
      SubValFromReg8(Register.A, Register.E)
      Register.PC += 1
    CASE &h94  ' SUB H
      PrintDebug("SUB H")  ' DEBUG
      SubValFromReg8(Register.A, Register.H)
      Register.PC += 1
    CASE &h95  ' SUB L
      PrintDebug("SUB L")  ' DEBUG
      SubValFromReg8(Register.A, Register.L)
      Register.PC += 1
    CASE &h96  ' SUB (HL)  (sub content at address HL from A)
      PrintDebug("SUB (HL)")  ' DEBUG
      SubValFromReg8(Register.A, MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &h97  ' SUB A
      PrintDebug("SUB A")  ' DEBUG
      SubValFromReg8(Register.A, Register.A)
      Register.PC += 1
    CASE &h98  ' SBC A,B
      PrintDebug("SBC A,B")  ' DEBUG
      SbcA(Register.B)
      Register.PC += 1
    CASE &h99  ' SBC A,C
      PrintDebug("SBC A,C")  ' DEBUG
      SbcA(Register.C)
      Register.PC += 1
    CASE &h9A  ' SBC A,D
      PrintDebug("SBC A,D")  ' DEBUG
      SbcA(Register.D)
      Register.PC += 1
    CASE &h9B  ' SBC A,E
      PrintDebug("SBC A,E")  ' DEBUG
      SbcA(Register.E)
      Register.PC += 1
    CASE &h9C  ' SBC A,H
      PrintDebug("SBC A,H")  ' DEBUG
      SbcA(Register.H)
      Register.PC += 1
    CASE &h9D  ' SBC A,L
      PrintDebug("SBC A,L")  ' DEBUG
      SbcA(Register.L)
      Register.PC += 1
    CASE &h9E  ' SBC A,(HL)
      PrintDebug("SBC A,(HL)")  ' DEBUG
      SbcA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &h9F  ' SBC A,A
      PrintDebug("SBC A,A")  ' DEBUG
      SbcA(Register.A)
      Register.PC += 1
    CASE &hA0   ' AND B (A = A AND B)
      PrintDebug("AND B")  ' DEBUG
      AndA(Register.B)
      Register.PC += 1
    CASE &hA1   ' AND C (A = A AND C)
      PrintDebug("AND C")  ' DEBUG
      AndA(Register.C)
      Register.PC += 1
    CASE &hA2   ' AND D (A = A AND D)
      PrintDebug("AND D")  ' DEBUG
      AndA(Register.D)
      Register.PC += 1
    CASE &hA3   ' AND E (A = A AND E)
      PrintDebug("AND E")  ' DEBUG
      AndA(Register.E)
      Register.PC += 1
    CASE &hA4   ' AND H (A = A AND H)
      PrintDebug("AND H")  ' DEBUG
      AndA(Register.H)
      Register.PC += 1
    CASE &hA5   ' AND L (A = A AND L)
      PrintDebug("AND L")  ' DEBUG
      AndA(Register.L)
      Register.PC += 1
    CASE &hA6   ' AND (HL) (A = A AND [HL])
      PrintDebug("AND (HL)")  ' DEBUG
      AndA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &hA7   ' AND A (A = A AND A)
      PrintDebug("AND A")  ' DEBUG
      AndA(Register.A)
      Register.PC += 1
    CASE &hA8   ' XOR B (A = A XOR B)
      PrintDebug("XOR B")  ' DEBUG
      XorA(Register.B) ' Xor with A, result in A
      Register.PC += 1
    CASE &hA9   ' XOR C (A = A XOR C)
      PrintDebug("XOR C")  ' DEBUG
      XorA(Register.C) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAA   ' XOR D (A = A XOR D)
      PrintDebug("XOR D")  ' DEBUG
      XorA(Register.D) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAB   ' XOR E (A = A XOR E)
      PrintDebug("XOR E")  ' DEBUG
      XorA(Register.E) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAC   ' XOR H (A = A XOR H)
      PrintDebug("XOR H")  ' DEBUG
      XorA(Register.H) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAD   ' XOR L (A = A XOR L)
      PrintDebug("XOR L")  ' DEBUG
      XorA(Register.L) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAE   ' XOR (HL)  (A = A XOR [HL])
      PrintDebug("XOR (HL)") ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      XorA(UbyteBuff) ' Xor with A, result in A
      Register.PC += 1
    CASE &hAF   ' XOR A (A = A XOR A)
      PrintDebug("XOR A")  ' DEBUG
      XorA(Register.A) ' Xor with A, result in A
      Register.PC += 1
    CASE &hB0   ' OR B  (A = A OR B)
      PrintDebug("OR B")   ' DEBUG
      OrA(Register.B)
      Register.PC += 1
    CASE &hB1   ' OR C  (A = A OR C)
      PrintDebug("OR C")   ' DEBUG
      OrA(Register.C)
      Register.PC += 1
    CASE &hB2   ' OR D  (A = A OR D)
      PrintDebug("OR D")   ' DEBUG
      OrA(Register.D)
      Register.PC += 1
    CASE &hB3   ' OR E  (A = A OR E)
      PrintDebug("OR E")   ' DEBUG
      OrA(Register.E)
      Register.PC += 1
    CASE &hB4   ' OR H  (A = A OR H)
      PrintDebug("OR H")   ' DEBUG
      OrA(Register.H)
      Register.PC += 1
    CASE &hB5   ' OR L  (A = A OR L)
      PrintDebug("OR L")   ' DEBUG
      OrA(Register.L)
      Register.PC += 1
    CASE &hB6   ' OR (HL)  (A = A OR (HL))
      PrintDebug("OR (HL)")   ' DEBUG
      OrA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &hB7   ' OR A  (A = A OR A)
      PrintDebug("OR A")   ' DEBUG
      OrA(Register.A)
      Register.PC += 1
    CASE &hB8   ' CP B (compare A with B)
      PrintDebug("CP B") ' DEBUG
      CmpA(Register.B)
      Register.PC += 1
    CASE &hB9   ' CP C (compare A with C)
      PrintDebug("CP C") ' DEBUG
      CmpA(Register.C)
      Register.PC += 1
    CASE &hBA   ' CP D (compare A with D)
      PrintDebug("CP D") ' DEBUG
      CmpA(Register.D)
      Register.PC += 1
    CASE &hBB   ' CP E (compare A with E)
      PrintDebug("CP E") ' DEBUG
      CmpA(Register.E)
      Register.PC += 1
    CASE &hBC   ' CP H (compare A with H)
      PrintDebug("CP H") ' DEBUG
      CmpA(Register.H)
      Register.PC += 1
    CASE &hBD   ' CP L (compare A with L)
      PrintDebug("CP L") ' DEBUG
      CmpA(Register.L)
      Register.PC += 1
    CASE &hBE   ' CP (HL) (compare A with value at address [HL])
      PrintDebug("CP (HL)") ' DEBUG
      CmpA(MemoryRead(ReadRegHL()))
      Register.PC += 1
    CASE &hBF   ' CP A (compare A with A)
      PrintDebug("CP A") ' DEBUG
      CmpA(Register.A)
      Register.PC += 1
    CASE &hC0   ' RET NZ   ' RET if Z = 0
      PrintDebug("RET NZ")
      IF GetFlagZ() = 0 THEN Register.PC = PopFromStack() ELSE Register.PC += 1
    CASE &hC1   ' POP BC (Write the value in stack into BC)
      PrintDebug("POP BC")  ' DEBUG
      WriteRegBC(PopFromStack())
      Register.PC += 1
    CASE &hC2   ' JP NZ,nn (jump if Z=0)
      PrintDebug("JP NZ," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Jump
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
    CASE &hC3   ' JP nn (Unconditional jump)
      PrintDebug("JP " & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
    CASE &hC4   ' CALL NZ,nn (call if Z=0)
      PrintDebug("CALL NZ," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Call
          PushToStack(Register.PC + 3)  ' First save the current state of the Program Counter (+3 bytes)
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
        ELSE   ' Don't call
          Register.PC += 3
      END IF
    CASE &hC5   ' PUSH BC
      PrintDebug("PUSH BC")  ' DEBUG
      PushToStack(ReadRegBC())
      Register.PC += 1
    CASE &hC6  ' ADD A,n  (A+=n)
      PrintDebug("ADD A," & HEX(MemoryRead(Register.PC + 1), 2))
      AddToA(MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hC7   ' RST 00h
      PrintDebug("RST 00h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h00  ' Jump to 00h
    CASE &hC8   ' RET Z   ' RET if Z = 1
      PrintDebug("RET Z")
      IF GetFlagZ() = 0 THEN Register.PC += 1 ELSE Register.PC = PopFromStack()
    CASE &hC9   ' RET
      PrintDebug("RET")
      Register.PC = PopFromStack()
    CASE &hCA   ' JP Z,nn (jump if Z=1)
      PrintDebug("JP Z," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagZ() = 0 THEN  ' Don't jump
          Register.PC += 3
        ELSE   ' Jump
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
      END IF
    CASE &hCB00  ' RLC B
      PrintDebug("RLC B")  ' DEBUG
      RotateRLC(Register.B)
      Register.PC += 2
    CASE &hCB01  ' RLC C
      PrintDebug("RLC C")  ' DEBUG
      RotateRLC(Register.C)
      Register.PC += 2
    CASE &hCB02  ' RLC D
      PrintDebug("RLC D")  ' DEBUG
      RotateRLC(Register.D)
      Register.PC += 2
    CASE &hCB03  ' RLC E
      PrintDebug("RLC E")  ' DEBUG
      RotateRLC(Register.E)
      Register.PC += 2
    CASE &hCB04  ' RLC H
      PrintDebug("RLC H")  ' DEBUG
      RotateRLC(Register.H)
      Register.PC += 2
    CASE &hCB05  ' RLC L
      PrintDebug("RLC L")  ' DEBUG
      RotateRLC(Register.L)
      Register.PC += 2
    CASE &hCB06  ' RLC (HL)
      PrintDebug("RLC (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      RotateRLC(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB07  ' RLC A
      PrintDebug("RLC A")  ' DEBUG
      RotateRLC(Register.A)
      Register.PC += 2
    CASE &hCB08  ' RRC B
      PrintDebug("RRC B")  ' DEBUG
      RotateRRC(Register.B)
      Register.PC += 2
    CASE &hCB09  ' RRC C
      PrintDebug("RRC C")  ' DEBUG
      RotateRRC(Register.C)
      Register.PC += 2
    CASE &hCB0A  ' RRC D
      PrintDebug("RRC D")  ' DEBUG
      RotateRRC(Register.D)
      Register.PC += 2
    CASE &hCB0B  ' RRC E
      PrintDebug("RRC E")  ' DEBUG
      RotateRRC(Register.E)
      Register.PC += 2
    CASE &hCB0C  ' RRC H
      PrintDebug("RRC H")  ' DEBUG
      RotateRRC(Register.H)
      Register.PC += 2
    CASE &hCB0D  ' RRC L
      PrintDebug("RRC L")  ' DEBUG
      RotateRRC(Register.L)
      Register.PC += 2
    CASE &hCB0E  ' RRC (HL)
      PrintDebug("RRC (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      RotateRRC(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB0F  ' RRC A
      PrintDebug("RRC A")  ' DEBUG
      RotateRRC(Register.A)
      Register.PC += 2
    CASE &hCB10  ' RL B
      PrintDebug("RL B")  ' DEBUG
      RotateLeftCarry(Register.B)
      Register.PC += 2
    CASE &hCB11  ' RL C
      PrintDebug("RL C")  ' DEBUG
      RotateLeftCarry(Register.C)
      Register.PC += 2
    CASE &hCB12  ' RL D
      PrintDebug("RL D")  ' DEBUG
      RotateLeftCarry(Register.D)
      Register.PC += 2
    CASE &hCB13  ' RL E
      PrintDebug("RL E")  ' DEBUG
      RotateLeftCarry(Register.E)
      Register.PC += 2
    CASE &hCB14  ' RL H
      PrintDebug("RL H")  ' DEBUG
      RotateLeftCarry(Register.H)
      Register.PC += 2
    CASE &hCB15  ' RL L
      PrintDebug("RL L")  ' DEBUG
      RotateLeftCarry(Register.L)
      Register.PC += 2
    CASE &hCB16  ' RL (HL)
      PrintDebug("RL (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      RotateLeftCarry(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB17  ' RL A
      PrintDebug("RL A")  ' DEBUG
      RotateLeftCarry(Register.A)
      Register.PC += 2
    CASE &hCB18  ' RR B
      PrintDebug("RR B")  ' DEBUG
      RotateRightCarry(Register.B)
      Register.PC += 2
    CASE &hCB19  ' RR C
      PrintDebug("RR C")  ' DEBUG
      RotateRightCarry(Register.C)
      Register.PC += 2
    CASE &hCB1A  ' RR D
      PrintDebug("RR D")  ' DEBUG
      RotateRightCarry(Register.D)
      Register.PC += 2
    CASE &hCB1B  ' RR E
      PrintDebug("RR E")  ' DEBUG
      RotateRightCarry(Register.E)
      Register.PC += 2
    CASE &hCB1C  ' RR H
      PrintDebug("RR H")  ' DEBUG
      RotateRightCarry(Register.H)
      Register.PC += 2
    CASE &hCB1D  ' RR L
      PrintDebug("RR L")  ' DEBUG
      RotateRightCarry(Register.L)
      Register.PC += 2
    CASE &hCB1E  ' RR (HL)
      PrintDebug("RR (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      RotateRightCarry(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB1F  ' RR A
      PrintDebug("RR A")  ' DEBUG
      RotateRightCarry(Register.A)
      Register.PC += 2
    CASE &hCB20  ' SLA B
      PrintDebug("SLA B")  ' DEBUG
      ShiftSLA(Register.B)
      Register.PC += 2
    CASE &hCB21  ' SLA C
      PrintDebug("SLA C")  ' DEBUG
      ShiftSLA(Register.C)
      Register.PC += 2
    CASE &hCB22  ' SLA D
      PrintDebug("SLA D")  ' DEBUG
      ShiftSLA(Register.D)
      Register.PC += 2
    CASE &hCB23  ' SLA E
      PrintDebug("SLA E")  ' DEBUG
      ShiftSLA(Register.E)
      Register.PC += 2
    CASE &hCB24  ' SLA H
      PrintDebug("SLA H")  ' DEBUG
      ShiftSLA(Register.H)
      Register.PC += 2
    CASE &hCB25  ' SLA L
      PrintDebug("SLA L")  ' DEBUG
      ShiftSLA(Register.L)
      Register.PC += 2
    CASE &hCB26  ' SLA (HL)
      PrintDebug("SLA (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      ShiftSLA(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB27  ' SLA A
      PrintDebug("SLA A")  ' DEBUG
      ShiftSLA(Register.A)
      Register.PC += 2
    CASE &hCB28  ' SRA B
      PrintDebug("SRA B")  ' DEBUG
      ShiftSRA(Register.B)
      Register.PC += 2
    CASE &hCB29  ' SRA C
      PrintDebug("SRA C")  ' DEBUG
      ShiftSRA(Register.C)
      Register.PC += 2
    CASE &hCB2A  ' SRA D
      PrintDebug("SRA D")  ' DEBUG
      ShiftSRA(Register.D)
      Register.PC += 2
    CASE &hCB2B  ' SRA E
      PrintDebug("SRA E")  ' DEBUG
      ShiftSRA(Register.E)
      Register.PC += 2
    CASE &hCB2C  ' SRA H
      PrintDebug("SRA H")  ' DEBUG
      ShiftSRA(Register.H)
      Register.PC += 2
    CASE &hCB2D  ' SRA L
      PrintDebug("SRA L")  ' DEBUG
      ShiftSRA(Register.L)
      Register.PC += 2
    CASE &hCB2E  ' SRA (HL)
      PrintDebug("SRA (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      ShiftSRA(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB2F  ' SRA A
      PrintDebug("SRA A")  ' DEBUG
      ShiftSRA(Register.A)
      Register.PC += 2
    CASE &hCB30  ' SWAP B (swaps high and low nibbles of B)
      PrintDebug("SWAP B")  ' DEBUG
      SwapUbyte(Register.B)
      Register.PC += 2
    CASE &hCB31  ' SWAP C (swaps high and low nibbles of C)
      PrintDebug("SWAP C")  ' DEBUG
      SwapUbyte(Register.C)
      Register.PC += 2
    CASE &hCB32  ' SWAP D (swaps high and low nibbles of D)
      PrintDebug("SWAP D")  ' DEBUG
      SwapUbyte(Register.D)
      Register.PC += 2
    CASE &hCB33  ' SWAP E (swaps high and low nibbles of E)
      PrintDebug("SWAP E")  ' DEBUG
      SwapUbyte(Register.E)
      Register.PC += 2
    CASE &hCB34  ' SWAP H (swaps high and low nibbles of H)
      PrintDebug("SWAP H")  ' DEBUG
      SwapUbyte(Register.H)
      Register.PC += 2
    CASE &hCB35  ' SWAP L (swaps high and low nibbles of L)
      PrintDebug("SWAP L")  ' DEBUG
      SwapUbyte(Register.L)
      Register.PC += 2
    CASE &hCB36  ' SWAP (HL) (swaps high and low nibbles of value at address HL)
      PrintDebug("SWAP (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      SwapUbyte(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB37  ' SWAP A (swaps high and low nibbles of A)
      PrintDebug("SWAP A")  ' DEBUG
      SwapUbyte(Register.A)
      Register.PC += 2
    CASE &hCB38  ' SRL B  (Shift Right B)
      PrintDebug("SRL B")  ' DEBUG
      ShiftRightSRL(Register.B)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB39  ' SRL C  (Shift Right C)
      PrintDebug("SRL C")  ' DEBUG
      ShiftRightSRL(Register.C)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3A  ' SRL D  (Shift Right D)
      PrintDebug("SRL D")  ' DEBUG
      ShiftRightSRL(Register.D)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3B  ' SRL E  (Shift Right E)
      PrintDebug("SRL E")  ' DEBUG
      ShiftRightSRL(Register.E)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3C  ' SRL H  (Shift Right H)
      PrintDebug("SRL H")  ' DEBUG
      ShiftRightSRL(Register.H)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3D  ' SRL L  (Shift Right L)
      PrintDebug("SRL L")  ' DEBUG
      ShiftRightSRL(Register.L)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3E  ' SRL (HL)  (Shift Right value at address HL)
      PrintDebug("SRL (HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      ShiftRightSRL(UbyteBuff)
      MemoryWrite(ReadRegHL(), UbyteBuff)   ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB3F  ' SRL A  (Shift Right A)
      PrintDebug("SRL A")  ' DEBUG
      ShiftRightSRL(Register.A)  ' Shift right and set some flags
      Register.PC += 2
    CASE &hCB40 ' BIT 0,B (Test bit 0 of register B)
      PrintDebug("BIT 0,B")  ' DEBUG
      TestBit(0, Register.B)
      Register.PC += 2
    CASE &hCB41 ' BIT 0,C (Test bit 0 of register C)
      PrintDebug("BIT 0,C")  ' DEBUG
      TestBit(0, Register.C)
      Register.PC += 2
    CASE &hCB42 ' BIT 0,D (Test bit 0 of register D)
      PrintDebug("BIT 0,D")  ' DEBUG
      TestBit(0, Register.D)
      Register.PC += 2
    CASE &hCB43 ' BIT 0,E (Test bit 0 of register E)
      PrintDebug("BIT 0,E")  ' DEBUG
      TestBit(0, Register.E)
      Register.PC += 2
    CASE &hCB44 ' BIT 0,H (Test bit 0 of register H)
      PrintDebug("BIT 0,H")  ' DEBUG
      TestBit(0, Register.H)
      Register.PC += 2
    CASE &hCB45 ' BIT 0,L (Test bit 0 of register L)
      PrintDebug("BIT 0,L")  ' DEBUG
      TestBit(0, Register.L)
      Register.PC += 2
    CASE &hCB46 ' BIT 0,(HL) (Test bit 0 of value at address HL)
      PrintDebug("BIT 0,(HL)")  ' DEBUG
      TestBit(0, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB47 ' BIT 0,A (Test bit 0 of register A)
      PrintDebug("BIT 0,A")  ' DEBUG
      TestBit(0, Register.A)
      Register.PC += 2
    CASE &hCB48 ' BIT 1,B (Test bit 1 of register B)
      PrintDebug("BIT 1,B")  ' DEBUG
      TestBit(1, Register.B)
      Register.PC += 2
    CASE &hCB49 ' BIT 1,C (Test bit 1 of register C)
      PrintDebug("BIT 1,C")  ' DEBUG
      TestBit(1, Register.C)
      Register.PC += 2
    CASE &hCB4A ' BIT 1,D (Test bit 1 of register D)
      PrintDebug("BIT 1,D")  ' DEBUG
      TestBit(1, Register.D)
      Register.PC += 2
    CASE &hCB4B ' BIT 1,E (Test bit 1 of register E)
      PrintDebug("BIT 1,E")  ' DEBUG
      TestBit(1, Register.E)
      Register.PC += 2
    CASE &hCB4C ' BIT 1,H (Test bit 1 of register H)
      PrintDebug("BIT 1,H")  ' DEBUG
      TestBit(1, Register.H)
      Register.PC += 2
    CASE &hCB4D ' BIT 1,L (Test bit 1 of register L)
      PrintDebug("BIT 1,L")  ' DEBUG
      TestBit(1, Register.L)
      Register.PC += 2
    CASE &hCB4E ' BIT 1,(HL) (Test bit 1 of value at address HL)
      PrintDebug("BIT 1,(HL)")  ' DEBUG
      TestBit(1, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB4F ' BIT 1,A (Test bit 1 of register A)
      PrintDebug("BIT 1,A")  ' DEBUG
      TestBit(1, Register.A)
      Register.PC += 2
    CASE &hCB54 ' BIT 2,H (Test bit 2 of register H)
      PrintDebug("BIT 2,H")  ' DEBUG
      TestBit(2, Register.H)
      Register.PC += 2
    CASE &hCB55 ' BIT 2,L (Test bit 2 of register L)
      PrintDebug("BIT 2,L")  ' DEBUG
      TestBit(2, Register.L)
      Register.PC += 2
    CASE &hCB56 ' BIT 2,(HL) (Test bit 2 of value at [HL])
      PrintDebug("BIT 2,(HL)")  ' DEBUG
      TestBit(2, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB57 ' BIT 2,A (Test bit 2 of register A)
      PrintDebug("BIT 2,A")  ' DEBUG
      TestBit(2, Register.A)
      Register.PC += 2
    CASE &hCB58 ' BIT 3,B (Test bit 3 of register B)
      PrintDebug("BIT 3,B")  ' DEBUG
      TestBit(3, Register.B)
      Register.PC += 2
    CASE &hCB59 ' BIT 3,C (Test bit 3 of register C)
      PrintDebug("BIT 3,C")  ' DEBUG
      TestBit(3, Register.C)
      Register.PC += 2
    CASE &hCB5A ' BIT 3,D (Test bit 3 of register D)
      PrintDebug("BIT 3,D")  ' DEBUG
      TestBit(3, Register.D)
      Register.PC += 2
    CASE &hCB5B ' BIT 3,E (Test bit 3 of register E)
      PrintDebug("BIT 3,E")  ' DEBUG
      TestBit(3, Register.E)
      Register.PC += 2
    CASE &hCB5C ' BIT 3,H (Test bit 3 of register H)
      PrintDebug("BIT 3,H")  ' DEBUG
      TestBit(3, Register.H)
      Register.PC += 2
    CASE &hCB5D ' BIT 3,L (Test bit 3 of register L)
      PrintDebug("BIT 3,L")  ' DEBUG
      TestBit(3, Register.L)
      Register.PC += 2
    CASE &hCB5E ' BIT 3,(HL) (Test bit 3 of value at [HL])
      PrintDebug("BIT 3,(HL)")  ' DEBUG
      TestBit(3, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB5F ' BIT 3,A (Test bit 3 of register A)
      PrintDebug("BIT 3,A")  ' DEBUG
      TestBit(3, Register.A)
      Register.PC += 2
    CASE &hCB60 ' BIT 4,B (Test bit 4 of register B)
      PrintDebug("BIT 4,B")  ' DEBUG
      TestBit(4, Register.B)
      Register.PC += 2
    CASE &hCB61 ' BIT 4,C (Test bit 4 of register C)
      PrintDebug("BIT 4,C")  ' DEBUG
      TestBit(4, Register.C)
      Register.PC += 2
    CASE &hCB62 ' BIT 4,D (Test bit 4 of register D)
      PrintDebug("BIT 4,D")  ' DEBUG
      TestBit(4, Register.D)
      Register.PC += 2
    CASE &hCB63 ' BIT 4,E (Test bit 4 of register E)
      PrintDebug("BIT 4,E")  ' DEBUG
      TestBit(4, Register.E)
      Register.PC += 2
    CASE &hCB64 ' BIT 4,H (Test bit 4 of register H)
      PrintDebug("BIT 4,H")  ' DEBUG
      TestBit(4, Register.H)
      Register.PC += 2
    CASE &hCB65 ' BIT 4,L (Test bit 4 of register L)
      PrintDebug("BIT 4,L")  ' DEBUG
      TestBit(4, Register.L)
      Register.PC += 2
    CASE &hCB66 ' BIT 4,(HL) (Test bit 4 of value at [HL])
      PrintDebug("BIT 4,(HL)")  ' DEBUG
      TestBit(4, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB67 ' BIT 4,A (Test bit 4 of register A)
      PrintDebug("BIT 4,A")  ' DEBUG
      TestBit(4, Register.A)
      Register.PC += 2
    CASE &hCB68 ' BIT 5,B (Test bit 5 of register B)
      PrintDebug("BIT 5,B")  ' DEBUG
      TestBit(5, Register.B)
      Register.PC += 2
    CASE &hCB69 ' BIT 5,C (Test bit 5 of register C)
      PrintDebug("BIT 5,C")  ' DEBUG
      TestBit(5, Register.C)
      Register.PC += 2
    CASE &hCB6A ' BIT 5,D (Test bit 5 of register D)
      PrintDebug("BIT 5,D")  ' DEBUG
      TestBit(5, Register.D)
      Register.PC += 2
    CASE &hCB6B ' BIT 5,E (Test bit 5 of register E)
      PrintDebug("BIT 5,E")  ' DEBUG
      TestBit(5, Register.E)
      Register.PC += 2
    CASE &hCB6C ' BIT 5,H (Test bit 5 of register H)
      PrintDebug("BIT 5,H")  ' DEBUG
      TestBit(5, Register.H)
      Register.PC += 2
    CASE &hCB6D ' BIT 5,L (Test bit 5 of register L)
      PrintDebug("BIT 5,L")  ' DEBUG
      TestBit(5, Register.L)
      Register.PC += 2
    CASE &hCB6E ' BIT 5,(HL) (Test bit 5 of value at [HL])
      PrintDebug("BIT 5,(HL)")  ' DEBUG
      TestBit(5, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB6F ' BIT 5,A (Test bit 5 of register A)
      PrintDebug("BIT 5,A")  ' DEBUG
      TestBit(5, Register.A)
      Register.PC += 2
    CASE &hCB70 ' BIT 6,B (Test bit 6 of register B)
      PrintDebug("BIT 6,B")  ' DEBUG
      TestBit(6, Register.B)
      Register.PC += 2
    CASE &hCB71 ' BIT 6,C (Test bit 6 of register C)
      PrintDebug("BIT 6,C")  ' DEBUG
      TestBit(6, Register.C)
      Register.PC += 2
    CASE &hCB72 ' BIT 6,D (Test bit 6 of register D)
      PrintDebug("BIT 6,D")  ' DEBUG
      TestBit(6, Register.D)
      Register.PC += 2
    CASE &hCB73 ' BIT 6,E (Test bit 6 of register E)
      PrintDebug("BIT 6,E")  ' DEBUG
      TestBit(6, Register.E)
      Register.PC += 2
    CASE &hCB74 ' BIT 6,H (Test bit 6 of register H)
      PrintDebug("BIT 6,H")  ' DEBUG
      TestBit(6, Register.H)
      Register.PC += 2
    CASE &hCB75 ' BIT 6,L (Test bit 6 of register L)
      PrintDebug("BIT 6,L")  ' DEBUG
      TestBit(6, Register.L)
      Register.PC += 2
    CASE &hCB76 ' BIT 6,(HL) (Test bit 6 of value at [HL])
      PrintDebug("BIT 6,(HL)")  ' DEBUG
      TestBit(6, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB77 ' BIT 6,A (Test bit 6 of register A)
      PrintDebug("BIT 6,A")  ' DEBUG
      TestBit(6, Register.A)
      Register.PC += 2
    CASE &hCB78 ' BIT 7,B (Test bit 7 of register B)
      PrintDebug("BIT 7,B")  ' DEBUG
      TestBit(7, Register.B)
      Register.PC += 2
    CASE &hCB79 ' BIT 7,C (Test bit 7 of register C)
      PrintDebug("BIT 7,C")  ' DEBUG
      TestBit(7, Register.C)
      Register.PC += 2
    CASE &hCB7A ' BIT 7,D (Test bit 7 of register D)
      PrintDebug("BIT 7,D")  ' DEBUG
      TestBit(7, Register.D)
      Register.PC += 2
    CASE &hCB7B ' BIT 7,E (Test bit 7 of register E)
      PrintDebug("BIT 7,E")  ' DEBUG
      TestBit(7, Register.E)
      Register.PC += 2
    CASE &hCB7C ' BIT 7,H (Test bit 7 of register H)
      PrintDebug("BIT 7,H")  ' DEBUG
      TestBit(7, Register.H)
      Register.PC += 2
    CASE &hCB7D ' BIT 7,L (Test bit 7 of register L)
      PrintDebug("BIT 7,L")  ' DEBUG
      TestBit(7, Register.L)
      Register.PC += 2
    CASE &hCB7E ' BIT 7,(HL) (Test bit 7 of value at address HL)
      PrintDebug("BIT 7,(HL)")  ' DEBUG
      TestBit(7, MemoryRead(ReadRegHL()))
      Register.PC += 2
    CASE &hCB7F ' BIT 7,A (Test bit 7 of register A)
      PrintDebug("BIT 7,A")  ' DEBUG
      TestBit(7, Register.A)
      Register.PC += 2
    CASE &hCB80 ' RES 0,B (Reset bit 0 of register B)
      PrintDebug("RES 0,B")  ' DEBUG
      Register.B = Register.B AND &b11111110
      Register.PC += 2
    CASE &hCB81 ' RES 0,C (Reset bit 0 of register C)
      PrintDebug("RES 0,C")  ' DEBUG
      Register.C = Register.C AND &b11111110
      Register.PC += 2
    CASE &hCB82 ' RES 0,D (Reset bit 0 of register D)
      PrintDebug("RES 0,D")  ' DEBUG
      Register.D = Register.D AND &b11111110
      Register.PC += 2
    CASE &hCB83 ' RES 0,E (Reset bit 0 of register E)
      PrintDebug("RES 0,E")  ' DEBUG
      Register.E = Register.E AND &b11111110
      Register.PC += 2
    CASE &hCB84 ' RES 0,H (Reset bit 0 of register H)
      PrintDebug("RES 0,H")  ' DEBUG
      Register.H = Register.H AND &b11111110
      Register.PC += 2
    CASE &hCB85 ' RES 0,L (Reset bit 0 of register L)
      PrintDebug("RES 0,L")  ' DEBUG
      Register.L = Register.L AND &b11111110
      Register.PC += 2
    CASE &hCB86 ' RES 0,(HL) (Reset bit 0 of value at [HL])
      PrintDebug("RES 0,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11111110
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB87 ' RES 0,A (Reset bit 0 of register A)
      PrintDebug("RES 0,A")  ' DEBUG
      Register.A = Register.A AND &b11111110
      Register.PC += 2
    CASE &hCB88 ' RES 1,B (Reset bit 1 of register B)
      PrintDebug("RES 1,B")  ' DEBUG
      Register.B = (Register.B AND &b11111101)
      Register.PC += 2
    CASE &hCB89 ' RES 1,C (Reset bit 1 of register C)
      PrintDebug("RES 1,C")  ' DEBUG
      Register.C = (Register.C AND &b11111101)
      Register.PC += 2
    CASE &hCB8A ' RES 1,D (Reset bit 1 of register D)
      PrintDebug("RES 1,D")  ' DEBUG
      Register.D = (Register.D AND &b11111101)
      Register.PC += 2
    CASE &hCB8B ' RES 1,E (Reset bit 1 of register E)
      PrintDebug("RES 1,E")  ' DEBUG
      Register.E = (Register.E AND &b11111101)
      Register.PC += 2
    CASE &hCB8C ' RES 1,H (Reset bit 1 of register H)
      PrintDebug("RES 1,H")  ' DEBUG
      Register.H = (Register.H AND &b11111101)
      Register.PC += 2
    CASE &hCB8D ' RES 1,L (Reset bit 1 of register L)
      PrintDebug("RES 1,L")  ' DEBUG
      Register.L = (Register.L AND &b11111101)
      Register.PC += 2
    CASE &hCB8E ' RES 1,(HL) (Reset bit 1 of value at [HL])
      PrintDebug("RES 1,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11111101
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB8F ' RES 1,A (Reset bit 1 of register A)
      PrintDebug("RES 1,A")  ' DEBUG
      Register.A = (Register.A AND &b11111101)
      Register.PC += 2
    CASE &hCB90 ' RES 2,B (Reset bit 2 of register B)
      PrintDebug("RES 2,B")  ' DEBUG
      Register.B = (Register.B AND &b11111011)
      Register.PC += 2
    CASE &hCB91 ' RES 2,C (Reset bit 2 of register C)
      PrintDebug("RES 2,C")  ' DEBUG
      Register.C = (Register.C AND &b11111011)
      Register.PC += 2
    CASE &hCB92 ' RES 2,D (Reset bit 2 of register D)
      PrintDebug("RES 2,D")  ' DEBUG
      Register.D = (Register.D AND &b11111011)
      Register.PC += 2
    CASE &hCB93 ' RES 2,E (Reset bit 2 of register E)
      PrintDebug("RES 2,E")  ' DEBUG
      Register.E = (Register.E AND &b11111011)
      Register.PC += 2
    CASE &hCB94 ' RES 2,H (Reset bit 2 of register H)
      PrintDebug("RES 2,H")  ' DEBUG
      Register.H = (Register.H AND &b11111011)
      Register.PC += 2
    CASE &hCB95 ' RES 2,L (Reset bit 2 of register L)
      PrintDebug("RES 2,L")  ' DEBUG
      Register.L = (Register.L AND &b11111011)
      Register.PC += 2
    CASE &hCB96 ' RES 2,(HL) (Reset bit 2 of value at [HL])
      PrintDebug("RES 2,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11111011
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB97 ' RES 2,A (Reset bit 2 of register A)
      PrintDebug("RES 2,A")  ' DEBUG
      Register.A = (Register.A AND &b11111011)
      Register.PC += 2
    CASE &hCB98 ' RES 3,B (Reset bit 3 of register B)
      PrintDebug("RES 3,B")  ' DEBUG
      Register.B = (Register.B AND &b11110111)
      Register.PC += 2
    CASE &hCB99 ' RES 3,C (Reset bit 3 of register C)
      PrintDebug("RES 3,C")  ' DEBUG
      Register.C = (Register.C AND &b11110111)
      Register.PC += 2
    CASE &hCB9A ' RES 3,D (Reset bit 3 of register D)
      PrintDebug("RES 3,D")  ' DEBUG
      Register.D = (Register.D AND &b11110111)
      Register.PC += 2
    CASE &hCB9B ' RES 3,E (Reset bit 3 of register E)
      PrintDebug("RES 3,E")  ' DEBUG
      Register.E = (Register.E AND &b11110111)
      Register.PC += 2
    CASE &hCB9C ' RES 3,H (Reset bit 3 of register H)
      PrintDebug("RES 3,H")  ' DEBUG
      Register.H = (Register.H AND &b11110111)
      Register.PC += 2
    CASE &hCB9D ' RES 3,L (Reset bit 3 of register L)
      PrintDebug("RES 3,L")  ' DEBUG
      Register.L = (Register.L AND &b11110111)
      Register.PC += 2
    CASE &hCB9E ' RES 3,(HL) (Reset bit 3 of value at [HL])
      PrintDebug("RES 3,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11110111
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCB9F ' RES 3,A (Reset bit 3 of register A)
      PrintDebug("RES 3,A")  ' DEBUG
      Register.A = (Register.A AND &b11110111)
      Register.PC += 2
    CASE &hCBA0 ' RES 4,B (Reset bit 4 of register B)
      PrintDebug("RES 4,B")  ' DEBUG
      Register.B = (Register.B AND &b11101111)
      Register.PC += 2
    CASE &hCBA1 ' RES 4,C (Reset bit 4 of register C)
      PrintDebug("RES 4,C")  ' DEBUG
      Register.C = (Register.C AND &b11101111)
      Register.PC += 2
    CASE &hCBA2 ' RES 4,D (Reset bit 4 of register D)
      PrintDebug("RES 4,D")  ' DEBUG
      Register.D = (Register.D AND &b11101111)
      Register.PC += 2
    CASE &hCBA3 ' RES 4,E (Reset bit 4 of register E)
      PrintDebug("RES 4,E")  ' DEBUG
      Register.E = (Register.E AND &b11101111)
      Register.PC += 2
    CASE &hCBA4 ' RES 4,H (Reset bit 4 of register H)
      PrintDebug("RES 4,H")  ' DEBUG
      Register.H = (Register.H AND &b11101111)
      Register.PC += 2
    CASE &hCBA5 ' RES 4,L (Reset bit 4 of register L)
      PrintDebug("RES 4,L")  ' DEBUG
      Register.L = (Register.L AND &b11101111)
      Register.PC += 2
    CASE &hCBA6 ' RES 4,(HL) (Reset bit 4 of value at [HL])
      PrintDebug("RES 4,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11101111
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCBA7 ' RES 4,A (Reset bit 4 of register A)
      PrintDebug("RES 4,A")  ' DEBUG
      Register.A = (Register.A AND &b11101111)
      Register.PC += 2
    CASE &hCBA8 ' RES 5,B (Reset bit 5 of register B)
      PrintDebug("RES 5,B")  ' DEBUG
      Register.B = (Register.B AND &b11011111)
      Register.PC += 2
    CASE &hCBA9 ' RES 5,C (Reset bit 5 of register C)
      PrintDebug("RES 5,C")  ' DEBUG
      Register.C = (Register.C AND &b11011111)
      Register.PC += 2
    CASE &hCBAA ' RES 5,D (Reset bit 5 of register D)
      PrintDebug("RES 5,D")  ' DEBUG
      Register.D = (Register.D AND &b11011111)
      Register.PC += 2
    CASE &hCBAB ' RES 5,E (Reset bit 5 of register E)
      PrintDebug("RES 5,E")  ' DEBUG
      Register.E = (Register.E AND &b11011111)
      Register.PC += 2
    CASE &hCBAC ' RES 5,H (Reset bit 5 of register H)
      PrintDebug("RES 5,H")  ' DEBUG
      Register.H = (Register.H AND &b11011111)
      Register.PC += 2
    CASE &hCBAD ' RES 5,L (Reset bit 5 of register L)
      PrintDebug("RES 5,L")  ' DEBUG
      Register.L = (Register.L AND &b11011111)
      Register.PC += 2
    CASE &hCBAE ' RES 5,(HL) (Reset bit 5 of value at [HL])
      PrintDebug("RES 5,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b11011111
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCBAF ' RES 5,A (Reset bit 5 of register A)
      PrintDebug("RES 5,A")  ' DEBUG
      Register.A = Register.A AND &b11011111
      Register.PC += 2
    CASE &hCBB0 ' RES 6,B (Reset bit 6 of register B)
      PrintDebug("RES 6,B")  ' DEBUG
      Register.B = Register.B AND &b10111111
      Register.PC += 2
    CASE &hCBB1 ' RES 6,C (Reset bit 6 of register C)
      PrintDebug("RES 6,C")  ' DEBUG
      Register.C = Register.C AND &b10111111
      Register.PC += 2
    CASE &hCBB2 ' RES 6,D (Reset bit 6 of register D)
      PrintDebug("RES 6,D")  ' DEBUG
      Register.D = Register.D AND &b10111111
      Register.PC += 2
    CASE &hCBB3 ' RES 6,E (Reset bit 6 of register E)
      PrintDebug("RES 6,E")  ' DEBUG
      Register.E = Register.E AND &b10111111
      Register.PC += 2
    CASE &hCBB4 ' RES 6,H (Reset bit 6 of register H)
      PrintDebug("RES 6,H")  ' DEBUG
      Register.H = Register.H AND &b10111111
      Register.PC += 2
    CASE &hCBB5 ' RES 6,L (Reset bit 6 of register L)
      PrintDebug("RES 6,L")  ' DEBUG
      Register.L = Register.L AND &b10111111
      Register.PC += 2
    CASE &hCBB6 ' RES 6,(HL) (Reset bit 6 of value at [HL])
      PrintDebug("RES 6,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b10111111
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCBB7 ' RES 6,A (Reset bit 6 of register A)
      PrintDebug("RES 6,A")  ' DEBUG
      Register.A = Register.A AND &b10111111
      Register.PC += 2
    CASE &hCBB8 ' RES 7,B (Reset bit 7 of register B)
      PrintDebug("RES 7,B")  ' DEBUG
      Register.B = Register.B AND &b01111111
      Register.PC += 2
    CASE &hCBB9 ' RES 7,C (Reset bit 7 of register C)
      PrintDebug("RES 7,C")  ' DEBUG
      Register.C = Register.C AND &b01111111
      Register.PC += 2
    CASE &hCBBA ' RES 7,D (Reset bit 7 of register D)
      PrintDebug("RES 7,D")  ' DEBUG
      Register.D = Register.D AND &b01111111
      Register.PC += 2
    CASE &hCBBB ' RES 7,E (Reset bit 7 of register E)
      PrintDebug("RES 7,E")  ' DEBUG
      Register.E = Register.E AND &b01111111
      Register.PC += 2
    CASE &hCBBC ' RES 7,H (Reset bit 7 of register H)
      PrintDebug("RES 7,H")  ' DEBUG
      Register.H = Register.H AND &b01111111
      Register.PC += 2
    CASE &hCBBD ' RES 7,L (Reset bit 7 of register L)
      PrintDebug("RES 7,L")  ' DEBUG
      Register.L = Register.L AND &b01111111
      Register.PC += 2
    CASE &hCBBE ' RES 7,(HL) (Reset bit 7 of value at [HL])
      PrintDebug("RES 7,(HL)")  ' DEBUG
      UbyteBuff = MemoryRead(ReadRegHL())
      UbyteBuff = UbyteBuff AND &b01111111
      MemoryWrite(ReadRegHL(), UbyteBuff)
      Register.PC += 2
    CASE &hCBBF ' RES 7,A (Reset bit 7 of register A)
      PrintDebug("RES 7,A")  ' DEBUG
      Register.A = Register.A AND &b01111111
      Register.PC += 2
    CASE &hCBC0  ' SET 0,B  (Set bit 0 of B)
      PrintDebug("SET 0,B")   ' DEBUG
      Register.B = Register.B OR &b00000001
      Register.PC += 2
    CASE &hCBC1  ' SET 0,C  (Set bit 0 of C)
      PrintDebug("SET 0,C")   ' DEBUG
      Register.C = Register.C OR &b00000001
      Register.PC += 2
    CASE &hCBC2  ' SET 0,D  (Set bit 0 of D)
      PrintDebug("SET 0,D")   ' DEBUG
      Register.D = Register.D OR &b00000001
      Register.PC += 2
    CASE &hCBC3  ' SET 0,E  (Set bit 0 of E)
      PrintDebug("SET 0,E")   ' DEBUG
      Register.E = Register.E OR &b00000001
      Register.PC += 2
    CASE &hCBC4  ' SET 0,H  (Set bit 0 of H)
      PrintDebug("SET 0,H")   ' DEBUG
      Register.H = Register.H OR &b00000001
      Register.PC += 2
    CASE &hCBC5  ' SET 0,L  (Set bit 0 of L)
      PrintDebug("SET 0,L")   ' DEBUG
      Register.L = Register.L OR &b00000001
      Register.PC += 2
    CASE &hCBC6  ' SET 0,(HL)  (Set bit 0 of value at HL)
      PrintDebug("SET 0,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00000001)
      Register.PC += 2
    CASE &hCBC7  ' SET 0,A  (Set bit 0 of A)
      PrintDebug("SET 0,A")   ' DEBUG
      Register.A = Register.A OR &b00000001
      Register.PC += 2
    CASE &hCBC8  ' SET 1,B  (Set bit 1 of B)
      PrintDebug("SET 1,B")   ' DEBUG
      Register.B = Register.B OR &b00000010
      Register.PC += 2
    CASE &hCBC9  ' SET 1,C  (Set bit 1 of C)
      PrintDebug("SET 1,C")   ' DEBUG
      Register.C = Register.C OR &b00000010
      Register.PC += 2
    CASE &hCBCA  ' SET 1,D  (Set bit 1 of D)
      PrintDebug("SET 1,D")   ' DEBUG
      Register.D = Register.D OR &b00000010
      Register.PC += 2
    CASE &hCBCB  ' SET 1,E  (Set bit 1 of E)
      PrintDebug("SET 1,E")   ' DEBUG
      Register.E = Register.E OR &b00000010
      Register.PC += 2
    CASE &hCBCC  ' SET 1,H  (Set bit 1 of H)
      PrintDebug("SET 1,H")   ' DEBUG
      Register.H = Register.H OR &b00000010
      Register.PC += 2
    CASE &hCBCD  ' SET 1,L  (Set bit 1 of L)
      PrintDebug("SET 1,L")   ' DEBUG
      Register.L = Register.L OR &b00000010
      Register.PC += 2
    CASE &hCBCE  ' SET 1,(HL)  (Set bit 1 of value at HL)
      PrintDebug("SET 1,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00000010)
      Register.PC += 2
    CASE &hCBCF  ' SET 1,A  (Set bit 1 of A)
      PrintDebug("SET 1,A")   ' DEBUG
      Register.A = Register.A OR &b00000010
      Register.PC += 2
    CASE &hCBD0  ' SET 2,B  (Set bit 2 of B)
      PrintDebug("SET 2,B")   ' DEBUG
      Register.B = Register.B OR &b00000100
      Register.PC += 2
    CASE &hCBD1  ' SET 2,C  (Set bit 2 of C)
      PrintDebug("SET 2,C")   ' DEBUG
      Register.C = Register.C OR &b00000100
      Register.PC += 2
    CASE &hCBD2  ' SET 2,D  (Set bit 2 of D)
      PrintDebug("SET 2,D")   ' DEBUG
      Register.D = Register.D OR &b00000100
      Register.PC += 2
    CASE &hCBD3  ' SET 2,E  (Set bit 2 of E)
      PrintDebug("SET 2,E")   ' DEBUG
      Register.E = Register.E OR &b00000100
      Register.PC += 2
    CASE &hCBD4  ' SET 2,H  (Set bit 2 of H)
      PrintDebug("SET 2,H")   ' DEBUG
      Register.H = Register.H OR &b00000100
      Register.PC += 2
    CASE &hCBD5  ' SET 2,L  (Set bit 2 of L)
      PrintDebug("SET 2,L")   ' DEBUG
      Register.L = Register.L OR &b00000100
      Register.PC += 2
    CASE &hCBD6  ' SET 2,(HL)  (Set bit 2 of value at HL)
      PrintDebug("SET 2,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00000100)
      Register.PC += 2
    CASE &hCBD7  ' SET 2,A  (Set bit 2 of A)
      PrintDebug("SET 2,A")   ' DEBUG
      Register.A = Register.A OR &b00000100
      Register.PC += 2
    CASE &hCBD8  ' SET 3,B  (Set bit 3 of B)
      PrintDebug("SET 3,B")   ' DEBUG
      Register.B = Register.B OR &b00001000
      Register.PC += 2
    CASE &hCBD9  ' SET 3,C  (Set bit 3 of C)
      PrintDebug("SET 3,C")   ' DEBUG
      Register.C = Register.C OR &b00001000
      Register.PC += 2
    CASE &hCBDA  ' SET 3,D  (Set bit 3 of D)
      PrintDebug("SET 3,D")   ' DEBUG
      Register.D = Register.D OR &b00001000
      Register.PC += 2
    CASE &hCBDB  ' SET 3,E  (Set bit 3 of E)
      PrintDebug("SET 3,E")   ' DEBUG
      Register.E = Register.E OR &b00001000
      Register.PC += 2
    CASE &hCBDC  ' SET 3,H  (Set bit 3 of H)
      PrintDebug("SET 3,H")   ' DEBUG
      Register.H = Register.H OR &b00001000
      Register.PC += 2
    CASE &hCBDD  ' SET 3,L  (Set bit 3 of L)
      PrintDebug("SET 3,L")   ' DEBUG
      Register.L = Register.L OR &b00001000
      Register.PC += 2
    CASE &hCBDE  ' SET 3,(HL)  (Set bit 3 of value at HL)
      PrintDebug("SET 3,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00001000)
      Register.PC += 2
    CASE &hCBDF  ' SET 3,A  (Set bit 3 of A)
      PrintDebug("SET 3,A")   ' DEBUG
      Register.A = Register.A OR &b00001000
      Register.PC += 2
    CASE &hCBE0  ' SET 4,B  (Set bit 4 of B)
      PrintDebug("SET 4,B")   ' DEBUG
      Register.B = Register.B OR &b00010000
      Register.PC += 2
    CASE &hCBE1  ' SET 4,C  (Set bit 4 of C)
      PrintDebug("SET 4,C")   ' DEBUG
      Register.C = Register.C OR &b00010000
      Register.PC += 2
    CASE &hCBE2  ' SET 4,D  (Set bit 4 of D)
      PrintDebug("SET 4,D")   ' DEBUG
      Register.D = Register.D OR &b00010000
      Register.PC += 2
    CASE &hCBE3  ' SET 4,E  (Set bit 4 of E)
      PrintDebug("SET 4,E")   ' DEBUG
      Register.E = Register.E OR &b00010000
      Register.PC += 2
    CASE &hCBE4  ' SET 4,H  (Set bit 4 of H)
      PrintDebug("SET 4,H")   ' DEBUG
      Register.H = Register.H OR &b00010000
      Register.PC += 2
    CASE &hCBE5  ' SET 4,L  (Set bit 4 of L)
      PrintDebug("SET 4,L")   ' DEBUG
      Register.L = Register.L OR &b00010000
      Register.PC += 2
    CASE &hCBE6  ' SET 4,(HL)  (Set bit 4 of value at HL)
      PrintDebug("SET 4,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00010000)
      Register.PC += 2
    CASE &hCBE7  ' SET 4,A  (Set bit 4 of A)
      PrintDebug("SET 4,A")   ' DEBUG
      Register.A = Register.A OR &b00010000
      Register.PC += 2
    CASE &hCBE8  ' SET 5,B  (Set bit 5 of B)
      PrintDebug("SET 5,B")   ' DEBUG
      Register.B = Register.B OR &b00100000
      Register.PC += 2
    CASE &hCBE9  ' SET 5,C  (Set bit 5 of C)
      PrintDebug("SET 5,C")   ' DEBUG
      Register.C = Register.C OR &b00100000
      Register.PC += 2
    CASE &hCBEA  ' SET 5,D  (Set bit 5 of D)
      PrintDebug("SET 5,D")   ' DEBUG
      Register.D = Register.D OR &b00100000
      Register.PC += 2
    CASE &hCBEB  ' SET 5,E  (Set bit 5 of E)
      PrintDebug("SET 5,E")   ' DEBUG
      Register.E = Register.E OR &b00100000
      Register.PC += 2
    CASE &hCBEC  ' SET 5,H  (Set bit 5 of H)
      PrintDebug("SET 5,H")   ' DEBUG
      Register.H = Register.H OR &b00100000
      Register.PC += 2
    CASE &hCBED  ' SET 5,L  (Set bit 5 of L)
      PrintDebug("SET 5,L")   ' DEBUG
      Register.L = Register.L OR &b00100000
      Register.PC += 2
    CASE &hCBEE  ' SET 5,(HL)  (Set bit 5 of value at HL)
      PrintDebug("SET 5,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b00100000)
      Register.PC += 2
    CASE &hCBEF  ' SET 5,A  (Set bit 5 of A)
      PrintDebug("SET 5,A")   ' DEBUG
      Register.A = Register.A OR &b00100000
      Register.PC += 2
    CASE &hCBF0  ' SET 6,B  (Set bit 6 of B)
      PrintDebug("SET 6,B")   ' DEBUG
      Register.B = Register.B OR &b01000000
      Register.PC += 2
    CASE &hCBF1  ' SET 6,C  (Set bit 6 of C)
      PrintDebug("SET 6,C")   ' DEBUG
      Register.C = Register.C OR &b01000000
      Register.PC += 2
    CASE &hCBF2  ' SET 6,D  (Set bit 6 of D)
      PrintDebug("SET 6,D")   ' DEBUG
      Register.D = Register.D OR &b01000000
      Register.PC += 2
    CASE &hCBF3  ' SET 6,E  (Set bit 6 of E)
      PrintDebug("SET 6,E")   ' DEBUG
      Register.E = Register.E OR &b01000000
      Register.PC += 2
    CASE &hCBF4  ' SET 6,H  (Set bit 6 of H)
      PrintDebug("SET 6,H")   ' DEBUG
      Register.H = Register.H OR &b01000000
      Register.PC += 2
    CASE &hCBF5  ' SET 6,L  (Set bit 6 of L)
      PrintDebug("SET 6,L")   ' DEBUG
      Register.L = Register.L OR &b01000000
      Register.PC += 2
    CASE &hCBF6  ' SET 6,(HL)  (Set bit 6 of value at HL)
      PrintDebug("SET 6,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b01000000)
      Register.PC += 2
    CASE &hCBF7  ' SET 6,A  (Set bit 6 of A)
      PrintDebug("SET 6,A")   ' DEBUG
      Register.A = Register.A OR &b01000000
      Register.PC += 2
    CASE &hCBF8  ' SET 7,B  (Set bit 7 of B)
      PrintDebug("SET 7,B")   ' DEBUG
      Register.B = Register.B OR &b10000000
      Register.PC += 2
    CASE &hCBF9  ' SET 7,C  (Set bit 7 of C)
      PrintDebug("SET 7,C")   ' DEBUG
      Register.C = Register.C OR &b10000000
      Register.PC += 2
    CASE &hCBFA  ' SET 7,D  (Set bit 7 of D)
      PrintDebug("SET 7,D")   ' DEBUG
      Register.D = Register.D OR &b10000000
      Register.PC += 2
    CASE &hCBFB  ' SET 7,E  (Set bit 7 of E)
      PrintDebug("SET 7,E")   ' DEBUG
      Register.E = Register.E OR &b10000000
      Register.PC += 2
    CASE &hCBFC  ' SET 7,H  (Set bit 7 of H)
      PrintDebug("SET 7,H")   ' DEBUG
      Register.H = Register.H OR &b10000000
      Register.PC += 2
    CASE &hCBFD  ' SET 7,L  (Set bit 7 of L)
      PrintDebug("SET 7,L")   ' DEBUG
      Register.L = Register.L OR &b10000000
      Register.PC += 2
    CASE &hCBFE  ' SET 7,(HL)  (Set bit 7 of value at HL)
      PrintDebug("SET 7,(HL)")   ' DEBUG
      MemoryWrite(ReadRegHL(), MemoryRead(ReadRegHL()) OR &b10000000)
      Register.PC += 2
    CASE &hCBFF  ' SET 7,A  (Set bit 7 of A)
      PrintDebug("SET 7,A")   ' DEBUG
      Register.A = Register.A OR &b10000000
      Register.PC += 2
    CASE &hCC   ' CALL Z,nn (jump if Z=1)
      PrintDebug("CALL Z," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagZ() = 0 THEN   ' Don't jump
          Register.PC += 3
        ELSE  ' Jump
          PushToStack(Register.PC + 3)  ' First save the current state of the Program Counter (+3 bytes)
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
      END IF
    CASE &hCD   ' CALL (calls a subroutine)
      PrintDebug("CALL " & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      PushToStack(Register.PC + 3)  ' First save the current state of the Program Counter (+3 bytes)
      Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
    CASE &hCE   ' ADC A,n
      PrintDebug("ADC A," & HEX(MemoryRead(Register.PC + 1),2))   ' DEBUG
      AdcA(MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hCF   ' RST 08h
      PrintDebug("RST 08h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h08  ' Jump to 08h
    CASE &hD0   ' RET NC   ' RET if C = 0
      PrintDebug("RET NC")
      IF GetFlagC() = 0 THEN Register.PC = PopFromStack() ELSE Register.PC += 1
    CASE &hD1   ' POP DE (Write the value in stack into DE)
      PrintDebug("POP DE")  ' DEBUG
      WriteRegDE(PopFromStack())
      Register.PC += 1
    CASE &hD2   ' JP NC,nn (jump if C=0)
      PrintDebug("JP NC," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagC() = 0 THEN  ' Jump
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
    CASE &hD4   ' CALL NC,nn (jump if Z=0)
      PrintDebug("CALL NC," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagC() = 0 THEN  ' Jump
          PushToStack(Register.PC + 3)  ' First save the current state of the Program Counter (+3 bytes)
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
    CASE &hD5   ' PUSH DE
      PrintDebug("PUSH DE")  ' DEBUG
      PushToStack(ReadRegDE())
      Register.PC += 1
    CASE &hD6  ' SUB n from A
      PrintDebug("SUB " & HEX(MemoryRead(Register.PC + 1)))  ' DEBUG
      SubValFromReg8(Register.A, MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hD7   ' RST 10h
      PrintDebug("RST 10h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h10  ' Jump to 10h
    CASE &hD8   ' RET C   ' RET if C = 1
      PrintDebug("RET C")
      IF GetFlagC() = 0 THEN Register.PC += 1 ELSE Register.PC = PopFromStack()
    CASE &hD9   ' RETI  (RET + EI)
      PrintDebug("RETI")   ' DEBUG
      Register.PC = PopFromStack()   ' RET (return from a call or int)
      InterruptsState = 1            ' EI (enable interrupts)
    CASE &hDA   ' JP C,nn (jump if C=1)
      PrintDebug("JP C," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagC() = 0 THEN  ' Don't jump
          Register.PC += 3
        ELSE   ' Jump
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
      END IF
    CASE &hDC   ' CALL C,nn (jump if Z=1)
      PrintDebug("CALL C," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))) ' DEBUG
      IF GetFlagC() = 1 THEN  ' Jump
          PushToStack(Register.PC + 3)  ' First save the current state of the Program Counter (+3 bytes)
          Register.PC = DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))
        ELSE   ' Don't jump
          Register.PC += 3
      END IF
    CASE &hDF   ' RST 18h
      PrintDebug("RST 18h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &hDF  ' Jump to DFh
    CASE &hE0   ' LD ($FF00+n),A
      PrintDebug("LD (" & HEX(&hFF00 + MemoryRead(Register.PC + 1), 4) & "),A")  ' DEBUG
      MemoryWrite(&hFF00 + MemoryRead(Register.PC + 1), Register.A)
      Register.PC += 2
    CASE &hE1   ' POP HL (Write the value in stack into HL)
      PrintDebug("POP HL")  ' DEBUG
      WriteRegHL(PopFromStack())
      Register.PC += 1
    CASE &hE2   ' LD (FF00h+C),A (put A into addr $FF00+C)
      PrintDebug("LD (" & HEX(&hFF00 + Register.C) & "),A")   ' DEBUG
      MemoryWrite(&hFF00 + Register.C, Register.A)
      Register.PC += 1
    CASE &hE5   ' PUSH HL
      PrintDebug("PUSH HL")  ' DEBUG
      PushToStack(ReadRegHL())
      Register.PC += 1
    CASE &hE6   ' A AND n
      PrintDebug("A AND " & MemoryRead(Register.PC + 1))
      AndA(MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hE9   ' JP (HL) / Jumps to address contained in HL (same as "LD PC,HL")
      PrintDebug("JP (HL)")  ' DEBUG
      Register.PC = ReadRegHL()
    CASE &hEA   ' LD (nn),A   ' Load value of A into address nn
      PrintDebug("LD (" & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2))) & "),A")  ' DEBUG
      MemoryWrite(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)), Register.A)
      Register.PC += 3
    CASE &hEE   ' XOR n  (A = A XOR n)
      UbyteBuff = MemoryRead(Register.PC + 1)
      PrintDebug("XOR " & HEX(UbyteBuff, 2)) ' DEBUG
      XorA(UbyteBuff) ' Xor with A, result in A
      Register.PC += 2
    CASE &hEF   ' RST 28h
      PrintDebug("RST 28h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h28  ' Jump to 28h
    CASE &hF0   ' LD A,($FF00+n)
      PrintDebug("LD A,(" & HEX(&hFF00 + MemoryRead(Register.PC + 1), 4) & ")")  ' DEBUG
      Register.A = MemoryRead(&hFF00 + MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hF1   ' POP AF (Write the value in stack into AF)
      PrintDebug("POP AF")  ' DEBUG
      WriteRegAF(PopFromStack())
      Register.PC += 1
    CASE &hF2   ' LD A,(FF00h+C)  (put value at address FF00+C into A)
      PrintDebug("LD A,(FF00+C)")    ' DEBUG
      Register.A = MemoryRead(&hFF00 + Register.C)
      Register.PC += 1
    CASE &hF3   ' DI (Disables Interrupts)
      PrintDebug("DI - Disables interrupts")  ' DEBUG
      InterruptsState = 0   ' Sets IME (Interrupts Master Enable) to 0
      Register.PC += 1
    CASE &hF5   ' PUSH AF
      PrintDebug("PUSH AF")  ' DEBUG
      PushToStack(ReadRegAF())
      Register.PC += 1
    CASE &hF6   ' OR n (A = A OR n)
      PrintDebug("OR " & HEX(MemoryRead(Register.PC + 1), 2))  ' DEBUG
      OrA(MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hF7   ' RST 30h
      PrintDebug("RST 30h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h30  ' Jump to 30h
    CASE &hF9   ' LD SP,HL  (put HL into SP)
      PrintDebug("LD SP,HL")
      Register.SP = ReadRegHL()
      Register.PC += 1
    CASE &hFA   ' LD A,nn  (put value from address nn to A)
      PrintDebug("LD A," & HEX(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)), 4))
      Register.A = MemoryRead(DwordVal(MemoryRead(Register.PC + 1), MemoryRead(Register.PC + 2)))
      Register.PC += 3
    CASE &hFB   ' EI (Enables Interrupts)
      PrintDebug("EI - Enables interrupts")  ' DEBUG
      InterruptsState = 1
      Register.PC += 1
    CASE &hFE   ' CP n (compare A with n)
      PrintDebug("CP A," & HEX(MemoryRead(Register.PC + 1), 2)) ' DEBUG
      CmpA(MemoryRead(Register.PC + 1))
      Register.PC += 2
    CASE &hFF   ' RST 38h
      PrintDebug("RST 38h")   ' DEBUG
      PushToStack(Register.PC + 1)  ' Save current address (+1)
      Register.PC = &h38  ' Jump to 38h
    CASE ELSE
      IF CpuInstruction > 255 THEN
          PRINT "Unknown opcode at "; HEX(Register.PC); ": "; HEX(CpuInstruction, 4)
        ELSE
          PRINT "Unknown opcode at "; HEX(Register.PC); ": "; HEX(CpuInstruction, 2)
      END IF
      QuitEmulator = 1
  END SELECT
  #IFDEF DEBUGMODE
    OldLine = CSRLIN
    LOCATE 35, 20: PRINT "A:  "; HEX(Register.A, 2);  ' DEBUG
    LOCATE 36, 20: PRINT "B:  "; HEX(Register.B, 2);  ' DEBUG
    LOCATE 37, 20: PRINT "C:  "; HEX(Register.C, 2);  ' DEBUG
    LOCATE 38, 20: PRINT "D:  "; HEX(Register.D, 2);  ' DEBUG
    LOCATE 39, 20: PRINT "E:  "; HEX(Register.E, 2);  ' DEBUG
    LOCATE 40, 20: PRINT "F:  "; HEX(Register.F, 2);  ' DEBUG
    LOCATE 41, 20: PRINT "H:  "; HEX(Register.H, 2);  ' DEBUG
    LOCATE 42, 20: PRINT "L:  "; HEX(Register.L, 2);  ' DEBUG
    LOCATE 43, 20: PRINT "AF: "; HEX(ReadRegAF(), 4); ' DEBUG
    LOCATE 44, 20: PRINT "BC: "; HEX(ReadRegBC(), 4); ' DEBUG
    LOCATE 45, 20: PRINT "DE: "; HEX(ReadRegDE(), 4); ' DEBUG
    LOCATE 46, 20: PRINT "HL: "; HEX(ReadRegHL(), 4); ' DEBUG
    LOCATE 47, 20: PRINT "SP: "; HEX(Register.SP, 4); ' DEBUG
    LOCATE 48, 20: PRINT "PC: "; HEX(Register.PC, 4); ' DEBUG
    LOCATE 49, 20: PRINT "z:  "; GetFlagZ();  ' DEBUG
    LOCATE 50, 20: PRINT "h:  "; GetFlagH();  ' DEBUG
    LOCATE 51, 20: PRINT "n:  "; GetFlagN();  ' DEBUG
    LOCATE 52, 20: PRINT "c:  "; GetFlagC();  ' DEBUG
    LOCATE 53, 20: PRINT "LY: "; HEX(IoRegisters(&hFF44), 2)  ' DEBUG
    LOCATE 54, 20: PRINT "LYC:"; HEX(IoRegisters(&hFF45), 2)  ' DEBUG
    LOCATE 55, 20: PRINT "IME:" & InterruptsState             ' DEBUG
    LOCATE OldLine
    IF DebugCpuPause = 1 THEN
        DO                                                      ' DEBUG
          SLEEP 10,1                                            ' DEBUG
          LastKey = INKEY
        LOOP UNTIL LEN(LastKey) > 0                             ' DEBUG
      ELSE
        LastKey = INKEY
    END IF
    IF LastKey = "p" THEN PRINT "*" : SLEEP                  ' DEBUG
  #ENDIF
  uTimer(cpc(CpuInstruction AND 255))            ' Update uTimer
  VideoSysUpdate(cpc(CpuInstruction AND 255))    ' Update Video subsystem
  AdjustTiming(cpc(CpuInstruction AND 255))      ' Slow down CPU emulation to 4MHz
  IF LastKey = CHR(27) THEN QuitEmulator = 1
END SUB

